@@ -1,7 +1,7 @@
-function summary = doctest_collect(what, directives, summary, recursive, verbose, depth, fid)
+function summary = doctest_collect(w, directives, summary, recursive, verbose, depth, fid)
 %DOCTEST_COLLECT  Find and run doctests.
 %
-%   The parameter WHAT is the name of a class, directory, function or filename:
+%   The input W is the name of a class, directory, function or filename:
 %   * For a directory, calls itself on the contents, recursively if
 %     RECURSIVE is true;
 %   * For a class, all methods are tested;
@@ -10,67 +10,88 @@ function summary = doctest_collect(what, directives, summary, recursive, verbose
 %%
 % Copyright (c) 2010 Thomas Grenfell Smith
 % Copyright (c) 2015 Michael Walter
-% Copyright (c) 2015-2019 Colin B. Macdonald
+% Copyright (c) 2015-2019, 2022 Colin B. Macdonald
 % Copyright (c) 2015 Oliver Heimlich
 % Copyright (C) 2018 Mike Miller
-% This is Free Software, BSD-3-Clause, see doctest.m for details.
+% SPDX-License-Identifier: BSD-3-Clause
 
 
 % TODO: methods('logical') octave/matlab differ: which behaviour do we want?
 % TODO: what about builtin "test" versus dir "test/"?  Do we prefer dir?
 
+if (isempty(w))
+  return
+end
+
 % determine type of target
 if is_octave()
-  % Note: ripe for refactoring once "exist(what, 'class')" works in Octave.
-  [~, ~, ext] = fileparts(what);
+  % Note: ripe for refactoring once "exist(w, 'class')" works in Octave.
+  [~, ~, ext] = fileparts(w);
   if any(strcmpi(ext, {'.texinfo' '.texi' '.txi' '.tex'}))
     type = 'texinfo';
-  elseif (strcmp (ext, '.oct') && exist (what) == 3)  % .oct explicitly
+  elseif (strcmp (ext, '.oct') && exist (w) == 3)  % .oct explicitly
     type = 'octfile';
-  elseif (exist (what) == 3)  % .oct/.mex
-    [~, what, ~] = fileparts (what);  % strip extension if present
+  elseif (exist (w) == 3)  % .oct/.mex
+    [~, w, ~] = fileparts (w);  % strip extension if present
     type = 'function';                % then access like any function
-  elseif (exist(what, 'file') && ~exist(what, 'dir')) || exist(what, 'builtin')
-    if (exist(['@' what], 'dir'))
-      % special case, e.g., @logical is class, logical is builtin
-      type = 'class';
-    else
-      type = 'function';
-    end
-  elseif (strcmp(what(1), '@'))
-    % comes after 'file' above for "doctest @class/method"
-    type = 'class';
-  elseif (exist(what, 'dir'))
-    type = 'dir';
-  elseif exist(what) == 2 || exist(what) == 103
-    % Notes:
-    %   * exist('@class', 'dir') only works if pwd is the parent of
-    %     '@class', having it in the path is not sufficient.
-    %   * Return 2 on Octave 3.8 and 103 on Octave 4.
-    type = 'class';
   else
-    % classdef classes are not detected by any of the above
+    type = 'unknown';
+  end
+
+  %% Let's see if its a class by checking if methods returns
+  % What about classdef in oct file above?  Should we do this even if
+  % type is 'octfile'?
+  if (strcmp (type, 'unknown'))
+    if (~ isempty (w) && strcmp (w(1), '@'))
+      temp = w(2:end);
+    else
+      temp = w;
+    end
     try
-      temp = methods(what);
+      temp = methods(temp);
       type = 'class';
     catch
       type = 'unknown';
     end
   end
+
+  if (strcmp (type, 'unknown'))
+    if (exist(w, 'dir'))
+      type = 'dir';
+    elseif (exist(w, 'file') || exist(w, 'builtin') || exist(w) == 103)
+      type = 'function';
+    else
+      type = 'unknown';
+    end
+  end
+
+  % This covers anything that we can get help from that wasn't covered above,
+  % notably "doctest class.method".  Quite possibly some other things too.
+  if (strcmp (type, 'unknown'))
+    try
+      if (~isempty (help (w)))
+        type = 'function';
+      end
+    catch
+      % no-op
+    end
+  end
+
 else % Matlab
-  if (strcmp(what(1), '@')) && ~isempty(methods(what(2:end)))
+  if (strcmp(w(1), '@')) && ~isempty(methods(w(2:end)))
     % covers "doctest @class", but not "doctest @class/method"
     type = 'class';
-  elseif ~isempty(methods(what))
+  elseif ~isempty(methods(w))
     % covers "doctest class"
     type = 'class';
-  elseif (exist(what, 'dir'))
+  elseif (exist(w, 'dir'))
     type = 'dir';
-  elseif exist(what, 'file') || exist(what, 'builtin');
+  elseif exist(w, 'file') || exist(w, 'builtin');
     type = 'function';
-  elseif ~isempty(help(what))
+  elseif ~isempty(help(w))
     % covers "doctest class.method" and "doctest class/method"
-    type = 'function'
+    % no try-catch needed as no error when w has no help
+    type = 'function';
   else
     type = 'unknown';
   end
@@ -82,23 +103,23 @@ end
 
 % Deal with directories
 if (strcmp(type, 'dir'))
-  if (strcmp(what, '.'))
+  if (strcmp(w, '.'))
     if (depth == 0)
       % cheap hack to not indent when calling "doctest ."
       depth = -1;
     end
   else
     spaces = repmat(' ', 1, 2*depth);
-    if (strcmp(what(end), filesep()))
+    if (strcmp(w(end), filesep()))
       slashchar = '';
     else
       slashchar = filesep();
     end
     if (verbose)
-      fprintf(fid, '%s%s%s\n', spaces, what, slashchar);
+      fprintf(fid, '%s%s%s\n', spaces, w, slashchar);
     end
   end
-  oldcwd = chdir(what);
+  oldcwd = chdir(w);
   files = dir('.');
   for i=1:numel(files)
     f = files(i).name;
@@ -139,23 +160,23 @@ end
 %   TARGETS(i).depth      How "deep" in a recursive traversal are we
 
 if strcmp(type, 'function')
-  target = collect_targets_function(what);
+  target = collect_targets_function(w);
   target.depth = depth;
   targets = [target];
 elseif strcmp(type, 'class')
-  targets = collect_targets_class(what, depth);
+  targets = collect_targets_class(w, depth);
 elseif strcmp (type, 'octfile')
-  targets = collect_targets_octfile (what, depth);
+  targets = collect_targets_octfile (w, depth);
 elseif strcmp(type, 'texinfo')
   target = struct();
-  target.name = what;
+  target.name = w;
   target.link = '';
   target.depth = depth;
-  [target.docstring, target.error] = parse_texinfo(fileread(what));
+  [target.docstring, target.error] = parse_texinfo(fileread(w));
   targets = [target];
 else
   target = struct();
-  target.name = what;
+  target.name = w;
   target.link = '';
   target.depth = depth;
   target.docstring = '';
@@ -239,31 +260,31 @@ end
 
 
 
-function target = collect_targets_function(what)
+function target = collect_targets_function(w)
   target = struct();
-  target.name = what;
+  target.name = w;
   if is_octave()
     target.link = '';
   else
-    target.link = sprintf('<a href="matlab:editorservices.openAndGoToLine(''%s'', 1);">%s</a>', which(what), what);
+    target.link = sprintf('<a href="matlab:editorservices.openAndGoToLine(''%s'', 1);">%s</a>', which(w), w);
   end
   [target.docstring, target.error] = extract_docstring(target.name);
 end
 
 
-function targets = collect_targets_class(what, depth)
-  if (strcmp(what(1), '@'))
+function targets = collect_targets_class(w, depth)
+  if (strcmp(w(1), '@'))
     % Octave methods('@foo') gives java error, Matlab just says "No methods"
-    what = what(2:end);
+    w = w(2:end);
   end
 
-  % TODO: workaround github.com/catch22/octave-doctest/issues/135 by
+  % workaround github.com/catch22/octave-doctest/issues/135 by
   % accessing all non-constructor method help text *before* "help obj"
-  if (is_octave ())
-    meths = methods (what);
+  if (is_octave () && compare_versions (OCTAVE_VERSION, '7.0.0', '<'))
+    meths = methods (w);
     for i=1:numel (meths)
-      if (~ strcmp (meths{i}, what))  % skip @obj/obj
-        name = sprintf ('@%s%s%s', what, filesep (), meths{i});
+      if (~ strcmp (meths{i}, w))  % skip @obj/obj
+        name = sprintf ('@%s%s%s', w, filesep (), meths{i});
         [docstring, format] = get_help_text (name);
       end
     end
@@ -272,26 +293,38 @@ function targets = collect_targets_class(what, depth)
   % First, "help class".  For classdef, this differs from "help class.class"
   % (general class help vs constructor help).  For old-style classes we will
   % probably end up testing the constructor twice but... meh.
-  target.name = what;
+  target.name = w;
   if is_octave()
     target.link = '';
   else
-    target.link = sprintf('<a href="matlab:editorservices.openAndGoToLine(''%s'', 1);">%s</a>', which(what), what);
+    target.link = sprintf('<a href="matlab:editorservices.openAndGoToLine(''%s'', 1);">%s</a>', which(w), w);
   end
   target.depth = depth;
   [target.docstring, target.error] = extract_docstring(target.name);
   targets = target;
 
   % Next, add targets for all class methods
-  meths = methods(what);
+  meths = methods(w);
+  meths = unique (meths);  % Issue #260
   for i=1:numel(meths)
     target = struct();
     if is_octave()
-      target.name = sprintf('@%s%s%s', what, filesep(), meths{i});
-      target.link = '';
+      if compare_versions (OCTAVE_VERSION, '7.0.0', '>=') && exist (w, "class") == 8
+        % classdef on newish Octave: use cls.method
+        if strcmp (meths{i}, w)
+          % TODO: gathering the ctor help fails https://savannah.gnu.org/bugs/?62803
+          continue
+        end
+        target.name = sprintf ('%s.%s', w, meths{i});
+        target.link = '';
+      else
+        % use @cls/method for old-style classes https://savannah.gnu.org/bugs/?61521
+        target.name = sprintf ('@%s%s%s', w, filesep (), meths{i});
+        target.link = '';
+      end
     else
-      target.name = sprintf('%s.%s', what, meths{i});
-      target.link = sprintf('<a href="matlab:editorservices.openAndGoToFunction(''%s'', ''%s'');">%s</a>', which(what), meths{i}, target.name);
+      target.name = sprintf('%s.%s', w, meths{i});
+      target.link = sprintf('<a href="matlab:editorservices.openAndGoToFunction(''%s'', ''%s'');">%s</a>', which(w), meths{i}, target.name);
     end
     target.depth = depth;
     [target.docstring, target.error] = extract_docstring(target.name);
@@ -477,7 +510,7 @@ function [docstring, error] = parse_texinfo(str)
       % First nonblank line starts with '>>': assume diary style.  However,
       % we strip @result and @print macros (TODO: perhaps unwisely?)
       L = strsplit (T{i}, '\n');
-      L = regexprep (L, '^(\s*)(?:⇒|=>|⊣|-\|)', '$1', 'once', 'lineanchors');
+      L = regexprep (L, '^(\s*)(?:⇒|=>|⊣|-\||error→|error->)', '$1', 'once', 'lineanchors');
       T{i} = strjoin (L, '\n');
       continue
     end
@@ -486,7 +519,7 @@ function [docstring, error] = parse_texinfo(str)
     % Hack: the @example block is commonly mis-used to store non-examples such as
     % diagrams or math.  Delete an example block that has no indicated output.
     % (Hard to leave for "later" as we don't keep track of @example blocks.)
-    R1 = regexp (T{i}, '^\s*(⇒|=>|⊣|-\|)', 'lineanchors');
+    R1 = regexp (T{i}, '^\s*(⇒|=>|⊣|-\||error→|error->)', 'lineanchors');
     R2 = regexp (T{i}, '(doctest:\s+-TEXINFO_SKIP_BLOCKS_WO_OUTPUT)');
     T{i} = regexprep (T{i}, '(doctest:\s+-TEXINFO_SKIP_BLOCKS_WO_OUTPUT)', '');
     if (isempty (R1) && isempty (R2))
@@ -500,7 +533,7 @@ function [docstring, error] = parse_texinfo(str)
     % Categorize input and output lines in the example using
     % @result and @print macros.  Everything else, including comment lines and
     % empty lines, is categorized as input (for now).
-    Linput = cellfun ('isempty', regexp (L, '^\s*(⇒|=>|⊣|-\|)', 'once'));
+    Linput = cellfun ('isempty', regexp (L, '^\s*(⇒|=>|⊣|-\||error→|error->)', 'once'));
 
     if (not (Linput (1)))
       error = 'no command: @result on first line?';
@@ -549,7 +582,7 @@ function [docstring, error] = parse_texinfo(str)
     % strip @result and @print macro output
     Loutput = not (Linput);
     L(Loutput) = regexprep (L(Loutput), ...
-                            '^(\s*)(?:⇒|=>|⊣|-\|)', ...
+                            '^(\s*)(?:⇒|=>|⊣|-\||error→|error->)', ...
                             '$1', ...
                             'once', 'lineanchors');
 
